@@ -1,5 +1,5 @@
 /*
-    Copyright (c) 2005-2017 Intel Corporation
+    Copyright (c) 2005-2019 Intel Corporation
 
     Licensed under the Apache License, Version 2.0 (the "License");
     you may not use this file except in compliance with the License.
@@ -12,10 +12,6 @@
     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     See the License for the specific language governing permissions and
     limitations under the License.
-
-
-
-
 */
 
 #ifndef __TBB__concurrent_queue_impl_H
@@ -124,8 +120,8 @@ private:
 
 #if _MSC_VER && !defined(__INTEL_COMPILER)
 // unary minus operator applied to unsigned type, result still unsigned
-#pragma warning( push )
-#pragma warning( disable: 4146 )
+// #pragma warning( push )
+// #pragma warning( disable: 4146 )
 #endif
 
 //! A queue using simple locking.
@@ -391,7 +387,7 @@ micro_queue_pop_finalizer<T>::~micro_queue_pop_finalizer() {
 }
 
 #if _MSC_VER && !defined(__INTEL_COMPILER)
-#pragma warning( pop )
+// #pragma warning( pop )
 #endif // warning 4146 is back
 
 template<typename T> class concurrent_queue_iterator_rep ;
@@ -514,7 +510,7 @@ concurrent_queue_base_v3<T>::concurrent_queue_base_v3() {
     __TBB_ASSERT( (size_t)&my_rep->head_counter % NFS_GetLineSize()==0, "alignment error" );
     __TBB_ASSERT( (size_t)&my_rep->tail_counter % NFS_GetLineSize()==0, "alignment error" );
     __TBB_ASSERT( (size_t)&my_rep->array % NFS_GetLineSize()==0, "alignment error" );
-    memset(my_rep,0,sizeof(concurrent_queue_rep<T>));
+    memset(static_cast<void*>(my_rep),0,sizeof(concurrent_queue_rep<T>));
     my_rep->item_size = item_size;
     my_rep->items_per_page = item_size<=  8 ? 32 :
                              item_size<= 16 ? 16 :
@@ -538,12 +534,12 @@ bool concurrent_queue_base_v3<T>::internal_try_pop( void* dst ) {
             // Queue had item with ticket k when we looked.  Attempt to get that item.
             ticket tk=k;
 #if defined(_MSC_VER) && defined(_Wp64)
-    #pragma warning (push)
-    #pragma warning (disable: 4267)
+    // #pragma warning (push)
+    // #pragma warning (disable: 4267)
 #endif
             k = r.head_counter.compare_and_swap( tk+1, tk );
 #if defined(_MSC_VER) && defined(_Wp64)
-    #pragma warning (pop)
+    // #pragma warning (pop)
 #endif
             if( k==tk )
                 break;
@@ -889,7 +885,7 @@ protected:
     //! Get size of queue
     ptrdiff_t __TBB_EXPORTED_METHOD internal_size() const;
 
-    //! Check if the queue is emtpy
+    //! Check if the queue is empty
     bool __TBB_EXPORTED_METHOD internal_empty() const;
 
     //! Set the queue capacity
