@@ -1,5 +1,5 @@
 /*
-    Copyright (c) 2005-2017 Intel Corporation
+    Copyright (c) 2005-2019 Intel Corporation
 
     Licensed under the Apache License, Version 2.0 (the "License");
     you may not use this file except in compliance with the License.
@@ -12,10 +12,6 @@
     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     See the License for the specific language governing permissions and
     limitations under the License.
-
-
-
-
 */
 
 #ifndef __TBB_task_H
@@ -49,7 +45,9 @@ namespace internal { //< @cond INTERNAL
         task* self;
         task& parent;
     public:
-        explicit allocate_additional_child_of_proxy( task& parent_ ) : self(NULL), parent(parent_) {}
+        explicit allocate_additional_child_of_proxy( task& parent_ ) : self(NULL), parent(parent_) {
+            suppress_unused_warning( self );
+        }
         task& __TBB_EXPORTED_METHOD allocate( size_t size ) const;
         void __TBB_EXPORTED_METHOD free( task& ) const;
     };
@@ -162,6 +160,12 @@ namespace internal {
         void __TBB_EXPORTED_METHOD free( task& ) const;
     };
 
+#if __TBB_PREVIEW_CRITICAL_TASKS
+    // TODO: move to class methods when critical task API becomes public
+    void make_critical( task& t );
+    bool is_critical( task& t );
+#endif
+
     //! Memory prefix to a task object.
     /** This class is internal to the library.
         Do not reference it directly, except within the library itself.
@@ -187,6 +191,10 @@ namespace internal {
         friend class internal::allocate_child_proxy;
         friend class internal::allocate_continuation_proxy;
         friend class internal::allocate_additional_child_of_proxy;
+#if __TBB_PREVIEW_CRITICAL_TASKS
+        friend void make_critical( task& );
+        friend bool is_critical( task& );
+#endif
 
 #if __TBB_TASK_ISOLATION
         //! The tag used for task isolation.
@@ -274,6 +282,10 @@ namespace internal {
 #if __TBB_TASK_PRIORITY
 namespace internal {
     static const int priority_stride_v4 = INT_MAX / 4;
+#if __TBB_PREVIEW_CRITICAL_TASKS
+    // TODO: move into priority_t enum when critical tasks become public feature
+    static const int priority_critical = priority_stride_v4 * 3 + priority_stride_v4 / 3 * 2;
+#endif
 }
 
 enum priority_t {
@@ -294,6 +306,7 @@ enum priority_t {
 
 class task_scheduler_init;
 namespace interface7 { class task_arena; }
+using interface7::task_arena;
 
 //! Used to form groups of tasks
 /** @ingroup task_scheduling
@@ -320,7 +333,7 @@ class task_group_context : internal::no_copy {
 private:
     friend class internal::generic_scheduler;
     friend class task_scheduler_init;
-    friend class interface7::task_arena;
+    friend class task_arena;
 
 #if TBB_USE_CAPTURED_EXCEPTION
     typedef tbb_exception exception_container_type;
@@ -420,12 +433,16 @@ private:
     intptr_t my_priority;
 #endif /* __TBB_TASK_PRIORITY */
 
+    //! Description of algorithm for scheduler based instrumentation.
+    internal::string_index my_name;
+
     //! Trailing padding protecting accesses to frequently used members from false sharing
     /** \sa _leading_padding **/
     char _trailing_padding[internal::NFS_MaxLineSize - 2 * sizeof(uintptr_t) - 2 * sizeof(void*)
 #if __TBB_TASK_PRIORITY
-                            - sizeof(intptr_t)
+                           - sizeof(intptr_t)
 #endif /* __TBB_TASK_PRIORITY */
+                           - sizeof(internal::string_index)
                           ];
 
 public:
@@ -461,7 +478,17 @@ public:
     task_group_context ( kind_type relation_with_parent = bound,
                          uintptr_t t = default_traits )
         : my_kind(relation_with_parent)
-        , my_version_and_traits(2 | t)
+        , my_version_and_traits(3 | t)
+        , my_name(internal::CUSTOM_CTX)
+    {
+        init();
+    }
+
+    // Custom constructor for instrumentation of tbb algorithm
+    task_group_context ( internal::string_index name )
+        : my_kind(bound)
+        , my_version_and_traits(3 | default_traits)
+        , my_name(name)
     {
         init();
     }
@@ -783,11 +810,24 @@ public:
 #if __TBB_TASK_PRIORITY
     //! Enqueue task for starvation-resistant execution on the specified priority level.
     static void enqueue( task& t, priority_t p ) {
-        __TBB_ASSERT( p == priority_low || p == priority_normal || p == priority_high, "Invalid priority level value" );
+#if __TBB_PREVIEW_CRITICAL_TASKS
+        __TBB_ASSERT(p == priority_low || p == priority_normal || p == priority_high
+                     || p == internal::priority_critical, "Invalid priority level value");
+#else
+        __TBB_ASSERT(p == priority_low || p == priority_normal || p == priority_high, "Invalid priority level value");
+#endif
         t.prefix().owner->enqueue( t, (void*)p );
     }
 #endif /* __TBB_TASK_PRIORITY */
 
+    //! Enqueue task in task_arena
+    //! The implementation is in task_arena.h
+    inline static void enqueue( task& t, task_arena& arena
+#if __TBB_TASK_PRIORITY
+        , priority_t p = priority_t(0)
+#endif
+    );
+
     //! The innermost task being executed or destroyed by the current thread at the moment.
     static task& __TBB_EXPORTED_FUNC self();
 
@@ -906,8 +946,19 @@ private:
     internal::task_prefix& prefix( internal::version_tag* = NULL ) const {
         return reinterpret_cast<internal::task_prefix*>(const_cast<task*>(this))[-1];
     }
+#if __TBB_PREVIEW_CRITICAL_TASKS
+    friend void internal::make_critical( task& );
+    friend bool internal::is_critical( task& );
+#endif
 }; // class task
 
+#if __TBB_PREVIEW_CRITICAL_TASKS
+namespace internal {
+inline void make_critical( task& t ) { t.prefix().extra_state |= 0x8; }
+inline bool is_critical( task& t ) { return bool((t.prefix().extra_state & 0x8) != 0); }
+} // namespace internal
+#endif /* __TBB_PREVIEW_CRITICAL_TASKS */
+
 //! task that does nothing.  Useful for synchronization.
 /** @ingroup task_scheduling */
 class empty_task: public task {
@@ -954,7 +1005,7 @@ public:
     //! Destroys the list, but does not destroy the task objects.
     ~task_list() {}
 
-    //! True if list if empty; false otherwise.
+    //! True if list is empty; false otherwise.
     bool empty() const {return !first;}
 
     //! Push task onto back of list.
