@@ -1,5 +1,5 @@
 /*
-    Copyright (c) 2005-2017 Intel Corporation
+    Copyright (c) 2005-2019 Intel Corporation
 
     Licensed under the Apache License, Version 2.0 (the "License");
     you may not use this file except in compliance with the License.
@@ -12,10 +12,6 @@
     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
     See the License for the specific language governing permissions and
     limitations under the License.
-
-
-
-
 */
 
 #include "tbb/tbb_stddef.h"
@@ -152,7 +148,7 @@ market& market::global_market ( bool is_public, unsigned workers_requested, size
 #endif /* __TBB_TASK_GROUP_CONTEXT */
         __TBB_InitOnce::add_ref();
         void* storage = NFS_Allocate(1, size, NULL);
-        memset( storage, 0, size );
+        memset( static_cast<void*>(storage), 0, size );
         // Initialize and publish global market
         m = new (storage) market( workers_soft_limit, workers_hard_limit, stack_size );
         if( is_public )
@@ -190,7 +186,7 @@ bool market::release ( bool is_public, bool blocking_terminate ) {
                 // Theoretically, new private references to the market can be added during waiting making it potentially
                 // endless.
                 // TODO: revise why the weak scheduler needs market's pointer and try to remove this wait.
-                // Note that the market should know about its schedulers for cancelation/exception/priority propagation,
+                // Note that the market should know about its schedulers for cancellation/exception/priority propagation,
                 // see e.g. task_group_context::cancel_group_execution()
                 while ( __TBB_load_with_acquire( my_public_ref_count ) == 1 && __TBB_load_with_acquire( my_ref_count ) > 1 )
                     __TBB_Yield();
@@ -329,7 +325,7 @@ void market::try_destroy_arena ( arena* a, uintptr_t aba_epoch ) {
     assert_market_valid();
 #if __TBB_TASK_PRIORITY
     // scan all priority levels, not only in [my_global_bottom_priority;my_global_top_priority]
-    // range, because arena to be destoyed can have no outstanding request for workers
+    // range, because arena to be destroyed can have no outstanding request for workers
     for ( int p = num_priority_levels-1; p >= 0; --p ) {
         priority_level_info &pl = my_priority_levels[p];
         arena_list_type &my_arenas = pl.arenas;
@@ -360,10 +356,10 @@ void market::try_destroy_arena ( arena* a, uintptr_t aba_epoch ) {
 }
 
 /** This method must be invoked under my_arenas_list_mutex. **/
-arena* market::arena_in_need ( arena_list_type &arenas, arena *&next ) {
+arena* market::arena_in_need ( arena_list_type &arenas, arena *hint ) {
     if ( arenas.empty() )
         return NULL;
-    arena_list_type::iterator it = next;
+    arena_list_type::iterator it = hint;
     __TBB_ASSERT( it != arenas.end(), NULL );
     do {
         arena& a = *it;
@@ -375,11 +371,9 @@ arena* market::arena_in_need ( arena_list_type &arenas, arena *&next ) {
 #endif
             ) {
             a.my_references += arena::ref_worker;
-            as_atomic(next) = &*it; // a subject for innocent data race under the reader lock
-            // TODO: rework global round robin policy to local or random to avoid this write
             return &a;
         }
-    } while ( it != next );
+    } while ( it != hint );
     return NULL;
 }
 
@@ -415,6 +409,16 @@ int market::update_allotment ( arena_list_type& arenas, int workers_demand, int
     return assigned;
 }
 
+/** This method must be invoked under my_arenas_list_mutex. **/
+bool market::is_arena_in_list( arena_list_type &arenas, arena *a ) {
+    if ( a ) {
+        for ( arena_list_type::iterator it = arenas.begin(); it != arenas.end(); ++it )
+            if ( a == &*it )
+                return true;
+    }
+    return false;
+}
+
 #if __TBB_TASK_PRIORITY
 inline void market::update_global_top_priority ( intptr_t newPriority ) {
     GATHER_STATISTIC( ++governor::local_scheduler_if_initialized()->my_counters.market_prio_switches );
@@ -432,21 +436,29 @@ inline void market::reset_global_priority () {
     update_global_top_priority(normalized_normal_priority);
 }
 
-arena* market::arena_in_need ( arena* prev_arena )
-{
-    suppress_unused_warning(prev_arena);
+arena* market::arena_in_need ( arena* prev_arena ) {
     if( as_atomic(my_total_demand) <= 0 )
         return NULL;
     arenas_list_mutex_type::scoped_lock lock(my_arenas_list_mutex, /*is_writer=*/false);
     assert_market_valid();
     int p = my_global_top_priority;
     arena *a = NULL;
-    do {
-        priority_level_info &pl = my_priority_levels[p];
+
+    // Checks if arena is alive or not
+    if ( is_arena_in_list( my_priority_levels[p].arenas, prev_arena ) ) {
+        a = arena_in_need( my_priority_levels[p].arenas, prev_arena );
+    }
+
+    while ( !a && p >= my_global_bottom_priority ) {
+        priority_level_info &pl = my_priority_levels[p--];
         a = arena_in_need( pl.arenas, pl.next_arena );
+        if ( a ) {
+            as_atomic(pl.next_arena) = a; // a subject for innocent data race under the reader lock
+            // TODO: rework global round robin policy to local or random to avoid this write
+        }
         // TODO: When refactoring task priority code, take into consideration the
         // __TBB_TRACK_PRIORITY_LEVEL_SATURATION sections from earlier versions of TBB
-    } while ( !a && --p >= my_global_bottom_priority );
+    }
     return a;
 }
 
@@ -681,43 +693,23 @@ void market::adjust_demand ( arena& a, int delta ) {
 
 void market::process( job& j ) {
     generic_scheduler& s = static_cast<generic_scheduler&>(j);
-    arena *a = NULL;
+    // s.my_arena can be dead. Don't access it until arena_in_need is called
+    arena *a = s.my_arena;
     __TBB_ASSERT( governor::is_set(&s), NULL );
-    enum {
-        query_interval = 1000,
-        first_interval = 1
-    };
-    for(int i = first_interval; ; i--) {
-        while ( (a = arena_in_need(a)) )
-        {
+
+    for (int i = 0; i < 2; ++i) {
+        while ( (a = arena_in_need(a)) ) {
             a->process(s);
-            i = first_interval;
+            a = NULL; // to avoid double checks in arena_in_need(arena*) for the same priority level
         }
         // Workers leave market because there is no arena in need. It can happen earlier than
         // adjust_job_count_estimate() decreases my_slack and RML can put this thread to sleep.
         // It might result in a busy-loop checking for my_slack<0 and calling this method instantly.
-        // first_interval>0 and the yield refines this spinning.
-        if( i > 0 )
+        // the yield refines this spinning.
+        if ( !i )
             __TBB_Yield();
-        else
-#if !__TBB_SLEEP_PERMISSION
-            break;
-#else
-        { // i == 0
-#if __TBB_TASK_PRIORITY
-            arena_list_type &al = my_priority_levels[my_global_top_priority].arenas;
-#else /* __TBB_TASK_PRIORITY */
-            arena_list_type &al = my_arenas;
-#endif /* __TBB_TASK_PRIORITY */
-            if( al.empty() ) // races if any are innocent TODO: replace by an RML query interface
-                break; // no arenas left, perhaps going to shut down
-            if( the_global_observer_list.ask_permission_to_leave() )
-                break; // go sleep
-            __TBB_Yield();
-            i = query_interval;
-        }
-#endif// !__TBB_SLEEP_PERMISSION
     }
+    
     GATHER_STATISTIC( ++s.my_counters.market_roundtrips );
 }
 
